# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Python bindings for Clifford synthesis module."""

from typing import Any, ClassVar, overload

from mqt.core.ir import QuantumComputation

class TargetMetric:
    __members__: ClassVar[dict[TargetMetric, int]] = ...  # read-only
    depth: ClassVar[TargetMetric] = ...
    gates: ClassVar[TargetMetric] = ...
    two_qubit_gates: ClassVar[TargetMetric] = ...

    @overload
    def __init__(self, value: int) -> None: ...
    @overload
    def __init__(self, arg0: str) -> None: ...
    @overload
    def __init__(self, arg0: TargetMetric) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class Verbosity:
    __members__: ClassVar[dict[Verbosity, int]] = ...  # read-only
    none: ClassVar[Verbosity] = ...
    fatal: ClassVar[Verbosity] = ...
    error: ClassVar[Verbosity] = ...
    warning: ClassVar[Verbosity] = ...
    info: ClassVar[Verbosity] = ...
    debug: ClassVar[Verbosity] = ...
    verbose: ClassVar[Verbosity] = ...

    @overload
    def __init__(self, value: int) -> None: ...
    @overload
    def __init__(self, arg0: str) -> None: ...
    @overload
    def __init__(self, arg0: Verbosity) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class SynthesisConfiguration:
    """Class representing the configuration for the Clifford synthesis tehcniques."""

    dump_intermediate_results: bool
    gate_limit_factor: float
    initial_timestep_limit: int
    intermediate_results_path: str
    minimize_gates_after_depth_optimization: bool
    minimize_gates_after_two_qubit_gate_optimization: bool
    solver_parameters: dict[str, bool | int | float | str]
    target_metric: TargetMetric
    try_higher_gate_limit_for_two_qubit_gate_optimization: bool
    use_maxsat: bool
    use_symmetry_breaking: bool
    verbosity: Verbosity
    heuristic: bool
    split_size: int
    linear_search: bool

    def __init__(self) -> None: ...
    def json(self) -> dict[str, Any]: ...

class SynthesisResults:
    """Class representing the results of the Clifford synthesis techniques."""
    def __init__(self) -> None: ...
    def sat(self) -> bool: ...
    def unsat(self) -> bool: ...
    @property
    def circuit(self) -> str: ...
    @property
    def depth(self) -> int: ...
    @property
    def gates(self) -> int: ...
    @property
    def runtime(self) -> float: ...
    @property
    def single_qubit_gates(self) -> int: ...
    @property
    def solver_calls(self) -> int: ...
    @property
    def tableau(self) -> str: ...
    @property
    def two_qubit_gates(self) -> int: ...

class Tableau:
    """Class representing a Clifford tableau."""
    @overload
    def __init__(self, n: int, include_stabilizers: bool = False) -> None: ...
    @overload
    def __init__(self, description: str) -> None: ...
    @overload
    def __init__(self, stabilizers: str, destabilizers: str) -> None: ...

class CliffordSynthesizer:
    """The main class for the Clifford synthesis techniques."""
    @overload
    def __init__(self, initial_tableau: Tableau, target_tableau: Tableau) -> None: ...
    @overload
    def __init__(self, target_tableau: Tableau) -> None: ...
    @overload
    def __init__(self, qc: QuantumComputation, use_destabilizers: bool) -> None: ...
    @overload
    def __init__(self, initial_tableau: Tableau, qc: QuantumComputation) -> None: ...
    def synthesize(self, config: SynthesisConfiguration = ...) -> None: ...
    @property
    def results(self) -> SynthesisResults: ...
    @property
    def result_circuit(self) -> QuantumComputation: ...



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Load a calibration for a superconducting architecture."""

from __future__ import annotations

from typing import TYPE_CHECKING

from qiskit.transpiler.target import Target

if TYPE_CHECKING:
    from ....sc import Architecture


def load_calibration(architecture: Architecture, calibration: str | Target | None = None) -> None:
    """Load a superconducting calibration from a string, BackendProperties, or Target.

    Args:
        architecture: The architecture to load the calibration into.
        calibration: The calibration to load.
    """
    if calibration is None:
        return

    if isinstance(calibration, str):
        architecture.load_properties(calibration)
    elif isinstance(calibration, Target):
        from .import_backend import import_target

        architecture.load_properties(import_target(calibration))
    else:  # pragma: no cover
        msg = f"Calibration type {type(calibration)} not supported."
        raise TypeError(msg)



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Python bindings module for MQT QMAP's Neutral Atom State Preparation."""

from typing import Any

from mqt.core.ir import QuantumComputation

class NAStatePreparationSolver:
    """The MQT QMAP's Neutral Atom State Preparation Solver.

    The neutral atom state preparation solver generates an optimal sequence of
    neutral atom operations for a given state preparation circuit.
    """

    def __init__(
        self,
        max_x: int,
        max_y: int,
        max_c: int,
        max_r: int,
        max_h_offset: int,
        max_v_offset: int,
        max_h_dist: int,
        max_v_dist: int,
        min_entangling_y: int,
        max_entangling_y: int,
    ) -> None:
        """Create a solver instance for the neutral atom state preparation problem.

        The solver is based on a 2D grid abstraction of the neutral atom quantum
        computer. The 2D plane is divided into interaction sites. Each interaction site
        is denoted by abstract x- and y-coordinates. The parameter `max_x` specifies the
        maximum x-coordinate, and `max_y` specifies the maximum y-coordinate. In the
        center of an interaction site, sits an SLM trap. Around that trap there are
        several possible discrete AOD positions arranged as a grid. The specific
        position of an atom within an interaction site is determined by the x- and
        y-offset from the SLM trap, i.e., those can be positive and negative. The
        maximum absolute value of the x- and y-offset is specified by `max_h_offset` and
        `max_v_offset`, respectively. Then, the parameter `max_c` specifies the maximum
        number of AOD columns, and `max_r` specifies the maximum number of AOD rows.
        Finally, in order to interact during a Rydberg stage, atoms must be located
        within a certain distance. The maximum horizontal and vertical distance between
        two atoms is specified by `max_h_dist` and `max_v_dist`, respectively. The
        parameter `min_entangling_y` specifies the minimum y-coordinate for entangling
        operations, and `max_entangling_y` specifies the maximum y-coordinate for
        entangling operations. Hence, y-coordinates outside of this range are located in
        the storage zone.

        Note:
            The solver can only handle a single storage zone below the entangling zone,
            i.e., in this case `min_entangling_y` must be zero and `max_entangling_y`
            must be less than `max_y`.

        Args:
            max_x: is the maximum discrete x-coordinate of the interaction sites
            max_y: is the maximum discrete y-coordinate of the interaction sites
            max_c: is the maximum number of AOD columns
            max_r: is the maximum number of AOD rows
            max_h_offset: is the maximum horizontal offset of the atoms
            max_v_offset: is the maximum vertical offset of the atoms
            max_h_dist: is the maximum horizontal distance between two atoms
            max_v_dist: is the maximum vertical distance between two atoms
            min_entangling_y: is the minimum y-coordinate for entangling operations
            max_entangling_y: is the maximum y-coordinate for entangling operations

        Raises:
            ValueError: if one of the parameters is invalid, e.g., is a negative
            value
        """

    class Result:
        """Neutral Atom State Preparation Solver Result."""
        def __init__(self) -> None:
            """Create a result object."""
        def json(self) -> dict[str, Any]:
            """Returns the result as a JSON string.

            Returns:
                the result as a JSON string
            """

    def solve(
        self,
        ops: list[tuple[int, int]],
        num_qubits: int,
        num_stages: int,
        num_transfers: int | None = ...,
        mind_ops_order: bool = ...,
        shield_idle_qubits: bool = ...,
    ) -> Result:
        """Solve the neutral atom state preparation problem.

        The solver generates an optimal sequence of neutral atom operations for a given
        state preparation circuit. The circuit is given as a list of operations, where
        each operation is a pair of qubits. The sequence is divided into stages. Each
        stage is either a Rydberg stage or a transfer stage. In a Rydberg stage,
        adjacent qubits in the entangling zone undergo an entangling gate. In a transfer
        stage, atoms can be stored from AOD into SLM traps and loaded from SLM traps
        into AOD. At the end of each stage, the atoms are shuttled to their next
        position. The number of stages is specified by `num_stages`. The number of
        transfers is fixed by `num_transfers` if give. If this parameter is not
        specified, then the solver will determine the optimal number of transfers. The
        parameter `mind_ops_order` specifies whether the order of the operations in the
        circuit should be preserved. The parameter `shield_idle_qubits` specifies
        whether idle qubits should be shielded from the entangling operations.

        Note:
            To retrieve the list of qubit pairs from a quantum circuit, use the function
            :func:`get_ops_for_solver`.


            The returned solver's result can either directly exported to the JSON format
            by calling the method :func:`json` on the result object or the result object
            can be passed to the function :func:`generate_code` to generate code
            consisting of neutral atom operations.

        Args:
            ops: is the list of operations in the circuit
            num_qubits: is the number of qubits in the circuit
            num_stages: is the number of stages in the sequence
            num_transfers: (optional) is the number of transfers in the sequence
            mind_ops_order: is True if the order of the operations should be
                preserved
            shield_idle_qubits: is True if idle qubits should be shielded

        Returns:
            the result of the solver

        Raises:
            ValueError: if one of the numeral parameters is invalid, e.g., is a
            negative value
        """

def get_ops_for_solver(
    qc: QuantumComputation,
    operation_type: str,
    num_controls: int,
    quiet: bool = ...,
) -> list[tuple[int, int]]:
    """Extract entangling operations as list of qubit pairs from the circuit.

    Note:
        This function can only extract qubit pairs of two-qubit operations.
        I.e., the operands of the operation plus the controls must be equal to two.

    Args:
        qc: is the quantum circuit
        operation_type: is the type of operation to extract, e.g., "z" for CZ
            gates
        num_controls: is the number of controls the operation acts on, e.g., 1
            for CZ gates
        quiet: if True, suppresses warning when the circuit contains operations
            other than the specified operation type

    Returns:
        list of qubit pairs

    Raises:
        ValueError: if the circuit contains operations other than the specified
            operation type and quiet is False
        ValueError: if the operation has more than two operands including
            controls
    """

def generate_code(
    qc: QuantumComputation,
    result: NAStatePreparationSolver.Result,
    min_atom_dist: int = ...,
    no_interaction_radius: int = ...,
    zone_dist: int = ...,
) -> str:
    """Generate code for the given circuit using the solver's result.

    Some parameters of the abstraction from the 2D grid used for the solver
    must be provided again.

    Args:
        qc: is the quantum circuit
        result: is the result of the solver
        min_atom_dist: is the minimum distance between atoms
        no_interaction_radius: is the radius around an atom where no other atom
            can be placed during an entangling operation that should not interact with the
            atom
        zone_dist: is the distance between zones, i.e., the minimal distance
            between two atoms in different zones

    Raises:
        ValueError: if one of the numeral parameters is invalid, e.g., is a
            negative value
    """



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Functionality for computing good subarchitectures for quantum circuit mapping.

This file implements the methods presented in https://arxiv.org/abs/2210.09321.
"""

from __future__ import annotations

import contextlib
import pickle
from itertools import combinations
from pathlib import Path
from typing import TYPE_CHECKING, Optional

import rustworkx as rx
import rustworkx.visualization as rxviz

from ...._compat.importlib import resources

if TYPE_CHECKING:
    from collections.abc import Iterable

    from matplotlib import figure
    from qiskit.providers import BackendV1, BackendV2

    from ...._compat.typing import TypeAlias
    from ....sc import Architecture


with contextlib.suppress(TypeError):
    Graph: TypeAlias = rx.PyGraph[int, Optional[int]]

PartialOrder: TypeAlias = dict[tuple[int, int], set[tuple[int, int]]]

# Architectures for which precomputed orderings are available
precomputed_backends = ["rigetti_16", "ibm_guadalupe_16"]


class SubarchitectureOrder:
    """Class representing the partial order of (sub)architectures."""

    inactive_color: str = "#1f78b4"
    active_color: str = "#faf18e"

    def __init__(self) -> None:
        """Initialize a partial order."""
        self.arch: Graph = rx.PyGraph()
        self.subarch_order: PartialOrder = {}
        self.desirable_subarchitectures: PartialOrder = {}
        self.isomorphisms: dict[tuple[int, int], dict[tuple[int, int], dict[int, int]]] = {}

        self.__compute_subarchs()
        self.__compute_subarch_order()
        self.__compute_desirable_subarchitectures()

    @classmethod
    def from_retworkx_graph(cls, graph: Graph) -> SubarchitectureOrder:
        """Construct the partial order from retworkx graph.

        Args:
            graph: retworkx graph representing the architecture.

        Returns:
            The resulting partial order.
        """
        so = SubarchitectureOrder()
        so.arch = graph

        so.__compute_subarchs()
        so.__compute_subarch_order()
        so.__compute_desirable_subarchitectures()
        return so

    @classmethod
    def from_coupling_map(cls, coupling_map: Iterable[tuple[int, int]]) -> SubarchitectureOrder:
        """Construct partial order from coupling map defined as set of tuples of connected qubits.

        Args:
            coupling_map: Iterable of tuples of connected qubits.

        Returns:
            The resulting partial order.
        """
        num_nodes = max(max(int(u), int(v)) for u, v in coupling_map)
        graph: Graph = rx.PyGraph()
        graph.add_nodes_from(list(range(num_nodes + 1)))
        graph.add_edges_from_no_data(list(coupling_map))

        return cls.from_retworkx_graph(graph)

    @classmethod
    def from_backend(cls, backend: BackendV1) -> SubarchitectureOrder:
        """Construct the partial order from a coupling map defined as a Qiskit backend.

        Args:
            backend: Qiskit backend.

        Returns:
            The resulting partial order.
        """
        coupling_map = [(c[0], c[1]) for c in backend.configuration().coupling_map]
        return cls.from_coupling_map(coupling_map)

    @classmethod
    def from_backend_v2(cls, backend: BackendV2) -> SubarchitectureOrder:
        """Construct the partial order from a coupling map defined as a Qiskit backend.

        Args:
            backend: Qiskit backend.

        Returns:
            The resulting partial order.
        """
        coupling_map = [(c[0], c[1]) for c in backend.coupling_map]
        return cls.from_coupling_map(coupling_map)

    @classmethod
    def from_qmap_architecture(cls, arch: Architecture) -> SubarchitectureOrder:
        """Construct the partial order from a QMAP :class:`~mqt.qmap.Architecture` object.

        Args:
            arch: QMAP architecture.

        Returns:
            The resulting partial order.
        """
        return cls.from_coupling_map(arch.coupling_map)

    @classmethod
    def from_library(cls, lib_name: str | Path) -> SubarchitectureOrder:
        """Construct the partial order from a stored library.

        Args:
            lib_name: Path to the library.

        Returns:
            The resulting partial order.
        """
        path = Path(lib_name).with_suffix(".pickle")
        with path.open("rb") as f:
            temp = pickle.load(f)  # noqa: S301

        so = SubarchitectureOrder()
        so.__dict__.update(temp.__dict__)

        return so

    @classmethod
    def from_string(cls, name: str) -> SubarchitectureOrder:
        """Construct the partial order from a library name.

        Args:
            name: Name of the library.

        Returns:
            The resulting partial order.
        """
        if name in precomputed_backends:
            ref = resources.files("mqt.qmap") / "libs" / (name + ".pickle")
            with resources.as_file(ref) as lib_path:
                return cls.from_library(lib_path)
        return SubarchitectureOrder()

    def optimal_candidates(self, nqubits: int) -> list[Graph]:
        """Return optimal subarchitecture candidate.

        Args:
            nqubits:
                size of circuit for which the optimal candidate should be given.

        Returns:
            List of optimal subarchitecture candidates for circuits of the given size.
        """
        if nqubits <= 0 or nqubits > self.arch.num_nodes():
            msg = "Number of qubits must not be smaller or equal 0 or larger then number of physical qubits of architecture."
            raise ValueError(msg)

        if nqubits == self.arch.num_nodes():
            return [self.arch]

        cands = self.__cand(nqubits)
        trans_ord = self.__transitive_closure(self.subarch_order)
        ref_ord = self.__reflexive_closure(trans_ord)

        opt_cands = set(ref_ord[next(iter(cands))])
        for cand in cands:
            opt_cands = opt_cands.intersection(set(ref_ord[cand]))

        ordered_cands = list(opt_cands)
        ordered_cands.sort()
        for cand in ordered_cands:
            opt_cands = opt_cands.difference(trans_ord[cand])

        return [self.sgs[n][i] for (n, i) in opt_cands]

    def covering(self, nqubits: int, size: int) -> list[Graph]:
        """Return covering for nqubit circuits.

        Args:
            nqubits:
                size of circuit for which the covering should be given.
            size:
                limit for the size of the covering.

        Returns:
            Subarchitecture covering for circuits of the given size.

        Note:
            A smaller covering might be found.
        """
        cov = self.__cand(nqubits)
        po_trans = self.__transitive_closure(self.subarch_order)
        ref_trans_po = self.__reflexive_closure(po_trans)
        queue = list({el for cand in cov for el in ref_trans_po[cand]})
        queue.sort(reverse=True)

        po_inv = self.__inverse_relation(po_trans)

        while len(cov) > size:
            d = queue.pop()
            cov_d = cov.intersection(po_inv[d])
            if len(cov_d) > 1:
                cov = cov.difference(cov_d)
                cov.add(d)

        return [self.sgs[n][i] for n, i in cov]

    def store_library(self, lib_name: str | Path) -> None:
        """Store ordering.

        Args:
            lib_name: Path to the library.
        """
        path = Path(lib_name).with_suffix(".pickle")
        with path.open("wb") as f:
            pickle.dump(self, f)

    def draw_subarchitecture(self, subarchitecture: Graph | tuple[int, int]) -> figure.Figure:
        """Create a matplotlib figure showing subarchitecture within the entire architecture.

        Nodes that are part of the subarchitecture are drawn yellow.
        Nodes that are not part of the subarchitecture are drawn blue.

        Args:
            subarchitecture: Subarchitecture to be drawn.

        Returns:
            Matplotlib figure.
        """
        if isinstance(subarchitecture, tuple):
            subarchitecture = self.sgs[subarchitecture[0]][subarchitecture[1]]
        colors = [SubarchitectureOrder.inactive_color for _ in range(self.arch.num_nodes())]
        for node in subarchitecture.nodes():
            colors[node] = SubarchitectureOrder.active_color
        return rxviz.mpl_draw(self.arch, node_color=colors)

    def draw_subarchitectures(self, subarchitectures: list[Graph] | list[tuple[int, int]]) -> list[figure.Figure]:
        """Create matplotlib figures showing subarchitectures within the entire architecture.

        For each subarchitecture one figure is drawn.
        Nodes that are part of the subarchitecture are drawn yellow.
        Nodes that are not part of the subarchitecture are drawn blue.

        Args:
            subarchitectures: Subarchitectures to be drawn.

        Returns:
            List of matplotlib figures.
        """
        return [self.draw_subarchitecture(subarchitecture) for subarchitecture in subarchitectures]

    def __compute_subarchs(self) -> None:
        """Compute all subarchitectures of the architecture."""
        self.sgs: list[list[Graph]] = [[] for i in range(self.arch.num_nodes() + 1)]

        for i in range(1, self.arch.num_nodes() + 1):
            node_combinations = combinations(range(self.arch.num_nodes()), i)
            for sg in (self.arch.subgraph(selected_nodes) for selected_nodes in node_combinations):
                if rx.is_connected(sg):
                    new_class = True
                    for g in self.sgs[i]:
                        if rx.is_isomorphic(g, sg):
                            new_class = False
                            break
                    if new_class:
                        self.sgs[i].append(sg)
        # init orders
        for n in range(self.arch.num_nodes() + 1):
            for i in range(len(self.sgs[n])):
                self.subarch_order[n, i] = set()
                self.desirable_subarchitectures[n, i] = set()
                self.isomorphisms[n, i] = {}

    def __compute_subarch_order(self) -> None:
        """Compute subarchitecture order."""
        for n, sgs_n in enumerate(self.sgs[:-1]):
            for i, sg in enumerate(sgs_n):
                for j, parent_sg in enumerate(self.sgs[n + 1]):
                    matcher = rx.graph_vf2_mapping(parent_sg, sg, subgraph=True)
                    for iso in matcher:
                        self.subarch_order[n, i].add((n + 1, j))
                        iso_rev = {val: key for key, val in iso.items()}
                        self.isomorphisms[n, i][n + 1, j] = iso_rev
                        break  # One isomorphism suffices

    def __complete_isos(self) -> None:
        """Complete isomorphisms."""
        for n in reversed(range(1, len(self.sgs[:-1]))):
            for i in range(len(self.sgs[n])):
                for _, i_prime in self.subarch_order[n, i]:
                    self.__combine_iso_with_parent(n, i, i_prime)

    def __combine_iso_with_parent(self, n: int, i: int, j: int) -> None:
        """Combine all isomorphisms from sgs[n][i] with those from sgs[n+1][j]."""
        first = self.isomorphisms[n, i][n + 1, j]
        for (row, k), second in self.isomorphisms[n + 1, j].items():
            self.isomorphisms[n, i][row, k] = SubarchitectureOrder.__combine_isos(first, second)

    @staticmethod
    def __combine_isos(first: dict[int, int], second: dict[int, int]) -> dict[int, int]:
        """Combine two isomorphisms."""
        combined = {}
        for src, img in first.items():
            combined[src] = second[img]
        return combined

    def __transitive_closure(self, po: PartialOrder) -> PartialOrder:
        """Compute transitive closure of partial order."""
        po_trans: PartialOrder = {(self.arch.num_nodes(), 0): set()}

        for n in reversed(range(1, len(self.sgs[:-1]))):
            for i in range(len(self.sgs[n])):
                new_rel = set(po[n, i])
                po_trans[n, i] = new_rel.copy()
                for n_prime, i_prime in po_trans[n, i]:
                    new_rel = new_rel.union(po_trans[n_prime, i_prime])
                po_trans[n, i] = new_rel

        return po_trans

    @classmethod
    def __reflexive_closure(cls, po: PartialOrder) -> PartialOrder:
        """Compute reflexive closure of partial order."""
        po_ref = {}
        for k, v in po.items():
            v_copy = v.copy()
            v_copy.add(k)
            po_ref[k] = v_copy
        return po_ref

    def __inverse_relation(self, po: PartialOrder) -> PartialOrder:
        """Compute inverse relation of partial order."""
        po_inv: PartialOrder = {}
        for n in range(self.arch.num_nodes() + 1):
            for i in range(len(self.sgs[n])):
                po_inv[n, i] = set()
        for k, v in po.items():
            for e in v:
                po_inv[e].add(k)
        return po_inv

    def __path_order_less(self, n: int, i: int, n_prime: int, i_prime: int) -> bool:
        """Check if sgs[n][i] is less than sgs[n_prime][i_prime] in the path order."""
        lhs = self.sgs[n][i]
        rhs = self.sgs[n_prime][i_prime]
        iso = self.isomorphisms[n, i][n_prime, i_prime]
        for v in range(lhs.num_nodes()):
            for w in range(lhs.num_nodes()):
                if v is w:
                    continue
                if (
                    rx.dijkstra_shortest_path_lengths(lhs, v, lambda _x: 1, goal=w)[w]
                    > rx.dijkstra_shortest_path_lengths(rhs, iso[v], lambda _x: 1, goal=iso[w])[iso[w]]
                ):
                    return True
        return False

    def __compute_desirable_subarchitectures(self) -> None:
        """Compute desirable subarchitectures."""
        self.__complete_isos()
        for n in reversed(range(1, len(self.sgs[:-1]))):
            for i in range(len(self.sgs[n])):
                val = self.isomorphisms[n, i]
                for n_prime, i_prime in val:
                    if self.__path_order_less(n, i, n_prime, i_prime):
                        self.desirable_subarchitectures[n, i].add((n_prime, i_prime))
                des = list(self.desirable_subarchitectures[n, i])
                des.sort()
                new_des: set[tuple[int, int]] = set()
                for j, (n_prime, i_prime) in enumerate(reversed(des)):
                    idx = len(des) - j - 1
                    if not any((n_prime, i_prime) in self.subarch_order[k] for k in des[:idx]):
                        new_des.add((n_prime, i_prime))

                self.desirable_subarchitectures[n, i] = new_des
                if len(self.desirable_subarchitectures[n, i]) == 0:
                    self.desirable_subarchitectures[n, i].add((n, i))
        self.desirable_subarchitectures[self.arch.num_nodes(), 0] = {(self.arch.num_nodes(), 0)}

    def __cand(self, nqubits: int) -> set[tuple[int, int]]:
        return {
            des for (n, i), desirables in self.desirable_subarchitectures.items() if n == nqubits for des in desirables
        }


def ibm_guadalupe_subarchitectures() -> SubarchitectureOrder:
    """Load the precomputed ibm guadalupe subarchitectures.

    Returns:
        The subarchitecture order for the ibm_guadalupe architecture.
    """
    ref = resources.files("mqt.qmap") / "libs" / "ibm_guadalupe_16.pickle"
    with resources.as_file(ref) as path:
        return SubarchitectureOrder.from_library(path)


def rigetti_16_subarchitectures() -> SubarchitectureOrder:
    """Load the precomputed rigetti subarchitectures.

    Returns:
        The subarchitecture order for the 16-qubit Rigetti architecture.
    """
    ref = resources.files("mqt.qmap") / "libs" / "rigetti_16.pickle"
    with resources.as_file(ref) as path:
        return SubarchitectureOrder.from_library(path)



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

if sys.version_info >= (3, 11):
    from typing import Self
elif TYPE_CHECKING:
    from typing_extensions import Self
else:
    Self = object


__all__ = ["Self", "TypeAlias"]


def __dir__() -> list[str]:
    return __all__
